import sys
import asyncio

import soyutnet
from soyutnet import SoyutNet
from soyutnet.constants import GENERIC_ID, GENERIC_LABEL


def main(w=2, graph_filename=""):
    async def scheduled():
        await asyncio.sleep(0.005 * w)
        soyutnet.terminate()

    net = SoyutNet()

    reg = net.PTRegistry()
    o1 = net.Observer(verbose=True)
    p1 = net.Place("p1", initial_tokens={GENERIC_LABEL: [GENERIC_ID] * w}, observer=o1)
    o2 = net.Observer(verbose=True)
    p2 = net.Place("p2", initial_tokens={GENERIC_LABEL: [GENERIC_ID] * 0}, observer=o2)
    t1 = net.Transition("t1")
    t2 = net.Transition("t2")
    """Define places and transitions (PTs)"""

    p1.connect(t1, weight=w).connect(p2, weight=w).connect(t2).connect(p1)
    """Connect PTs"""

    reg.register(p1)
    reg.register(p2)
    reg.register(t1)
    reg.register(t2)
    """Save to a list of PTs"""

    soyutnet.run(reg, extra_routines=[scheduled()])
    print("Simulation is terminated.")

    records = reg.get_merged_records()
    for rec in records:
        net.print(rec)

    if graph_filename:
        with open(graph_filename, "w") as fh:
            fh.write(reg.generate_graph(label_names={GENERIC_LABEL: "@"}))

    return records


if __name__ == "__main__":
    main(int(sys.argv[1]), sys.argv[2] if len(sys.argv) > 2 else "")



import asyncio
from typing import (
    Any,
    Tuple,
    Awaitable,
    Callable,
)

from .constants import *
from .pt_common import PTCommon, Arc
from .observer import Observer


class Place(PTCommon):
    """
    Defines PTNet places.
    """

    def __init__(
        self,
        name: str = "",
        observer: Observer | None = None,
        observer_record_limit: int = 0,
        observer_verbose: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Constructor

        :param name: Name of the place.
        :param observer: Observer instance assigned to the place.
        :param observer_record_limit: Maximum number of records that is recorded by the observer. It is unlimited when chosen ``0``.
        :param observer_verbose: If set, observer will print new records when saved.
        """
        super().__init__(name=name, **kwargs)
        self._observer: Observer
        if observer is not None:
            self._observer = observer
            self._observer._set_place(self)

    async def _observe(self, requester: str = "") -> None:
        """
        Save token counts for each label to the observer's records.

        Also, add the number of tokens sent to the caller arc.

        :param token_count_in_arc: Number of tokens in the output arc of places must also be added to the count.
        """
        if self._observer is not None:
            await self._observer.save(requester=requester)


class SpecialPlace(Place):
    """
    Custom place class whose token processing methods can be overriden.
    """

    def __init__(
        self,
        name: str = "",
        consumer: Callable[["SpecialPlace"], Awaitable[None]] | None = None,
        producer: Callable[["SpecialPlace"], Awaitable[list[TokenType]]] | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Constructor.

        :param name: Name of the place.
        :param consumer: Custom :py:func:`soyutnet.pt_common.PTCommon._process_input_arcs` function.
        :param producer: Custom :py:func:`soyutnet.pt_common.PTCommon._process_output_arcs` function.
        """
        super().__init__(name=name, **kwargs)
        self._consumer: Callable[["SpecialPlace"], Awaitable[None]] | None = consumer
        """Custom :py:func:`soyutnet.pt_common.PTCommon._process_input_arcs` function."""
        self._producer: (
            Callable[["SpecialPlace"], Awaitable[list[TokenType]]] | None
        ) = producer
        """Custom :py:func:`soyutnet.pt_common.PTCommon._process_output_arcs` function."""

    async def _process_input_arcs(self) -> bool:
        """
        Calls custom producer function after the default
        :py:func:`soyutnet.pt_common.PTCommon._process_input_arcs`.

        :return: If ``True`` continues to processing tokens and output arcs, else loops back to processing input arcs.
        """
        result: bool = await super()._process_input_arcs()
        if self._producer is not None:
            tokens: list[TokenType] = await self._producer(self)
            if tokens:
                for token in tokens:
                    label: label_t = token[0]
                    count: int = self._put_token(token, strict=False)
                    if self._observer is not None:
                        await self._observer.inc_token_count(label)

                return True

        return result

    async def _process_output_arcs(self) -> None:
        """
        Handles consumer function first, if it is defined.

        See, :py:func:`soyutnet.place.Place._process_output_arcs`.
        """
        if self._consumer is not None:
            await self._consumer(self)

        await super()._process_output_arcs()

    async def _process_tokens(self) -> bool:
        """
        Calls custom consumer function. If it is ``None`` calls :py:func:`soyutnet.pt_common.PTCommon._process_tokens` function.
        """
        if not await super()._process_tokens():
            return False

        if self._observer is not None:
            await self._observe(self._name)

        return True

    async def observe(self, requester: str = "") -> None:
        return



import sys
import asyncio

import soyutnet
from soyutnet import SoyutNet
from soyutnet.constants import GENERIC_ID, GENERIC_LABEL, TokenType


def co_begin(action_count):
    token_ids = list(range(1, action_count + 1))

    async def producer(place):
        try:
            id: id_t = token_ids.pop(0)
            return [(GENERIC_LABEL, id)]
        except IndexError:
            pass

        return []

    output_token_ids = list(token_ids)

    async def consumer(place):
        token = place.get_token(GENERIC_LABEL)
        if not token:
            return
        output_token_ids.remove(token[1])
        if not output_token_ids:
            await net.sleep(0.01)
            soyutnet.terminate()

    net = SoyutNet()
    o1 = net.ComparativeObserver(expected={1: [((GENERIC_LABEL, action_count),)]})
    o31 = net.ComparativeObserver(
        expected={1: [((GENERIC_LABEL, 0),)] * action_count + [((GENERIC_LABEL, 1),)]}
    )
    reg = net.PTRegistry()
    p1 = net.Place("p1", initial_tokens={GENERIC_LABEL: token_ids})
    p2 = net.Place("p2", observer=o1)
    t1 = net.Transition("t1", record_firing=True)
    t2 = net.Transition("t2", record_firing=True)
    p1.connect(t1).connect(p2).connect(t2, weight=action_count)
    reg.register(p1)
    reg.register(p2)
    reg.register(t1)
    reg.register(t2)
    for i in range(action_count):
        obsv = None
        if i == 0:
            obsv = o31
        p3i = net.SpecialPlace(f"p3_{i}", consumer=consumer, observer=obsv)
        reg.register(p3i)
        t2.connect(p3i)

    total_firings = [0, 0]

    async def firing_test1():
        nonlocal total_firings
        while True:
            await t1.wait_for_firing()
            total_firings[0] += 1

    async def firing_test2():
        nonlocal total_firings
        while True:
            await t2.wait_for_firing()
            total_firings[1] += 1

    soyutnet.run(reg, extra_routines=[firing_test1(), firing_test2()])

    print(total_firings, t2.get_no_of_times_enabled())
    assert total_firings[0] == t1.get_no_of_times_enabled() - 1
    assert total_firings[0] == len(t1.get_firing_records()) - 1
    """t1 notifies before firing_test1 starts waiting misses waiting for the first one."""
    assert total_firings[1] == t2.get_no_of_times_enabled()
    assert total_firings[1] == len(t2.get_firing_records())


def co_end(action_count):
    async def producer(place):
        net.DEBUG(f"{place.ident()}")
        place_id = place.ident().split(",")[0].split("_")[1]
        try:
            return [(GENERIC_LABEL, int(place_id))]
        except IndexError:
            pass

        return []

    output_token_ids = list(range(1, action_count + 1))

    async def consumer(place):
        token = place.get_token(GENERIC_LABEL)
        if not token:
            return
        output_token_ids.remove(token[1])
        net.DEBUG(f"Token '{token}' is removed.")
        if not output_token_ids:
            soyutnet.terminate()

    net = SoyutNet()
    reg = net.PTRegistry()
    p1 = net.SpecialPlace("p0", consumer=consumer)
    t0 = net.Transition("t0")
    t0.connect(p1)
    reg.register(p1)
    reg.register(t0)
    for i in range(action_count):
        p0i = net.SpecialPlace(f"p0_{i+1}", producer=producer)
        reg.register(p0i)
        p0i.connect(t0)

    try:
        asyncio.run(soyutnet.main(reg))
    except asyncio.exceptions.CancelledError:
        pass


def sync_by_signal():
    token_ids = list(range(1, 11)) + [GENERIC_ID] * 30

    async def producer(place):
        try:
            id = token_ids.pop(0)
            net.DEBUG(f"Produced '{(GENERIC_LABEL, id)}'")
            return [(GENERIC_LABEL, id)]
        except IndexError:
            pass

        return []

    output_token_ids = token_ids[:10]

    async def consumer(place):
        token = place.get_token(GENERIC_LABEL)
        if not token:
            return
        if token[1] == GENERIC_ID:
            return
        output_token_ids.remove(token[1])
        net.DEBUG(f"Consumed '{token}'")
        if not output_token_ids:
            soyutnet.terminate()

    activate = False

    async def processor(place):
        nonlocal activate
        if place._get_token_count(GENERIC_LABEL) > 3:
            net.DEBUG(f"Activated. 'p4' will start consuming.")
            activate = True

        return activate

    net = SoyutNet()
    p0 = net.SpecialPlace("p0", producer=producer)
    p1 = net.SpecialPlace("p1", consumer=consumer)
    p2 = net.Place("p2", processor=processor)
    p3 = net.SpecialPlace("p3", producer=producer)
    p4 = net.SpecialPlace("p4", consumer=consumer)
    t0 = net.Transition("t0")
    t1 = net.Transition("t1")

    reg = net.PTRegistry()
    reg.register(p0)
    reg.register(p1)
    reg.register(p2)
    reg.register(p3)
    reg.register(p4)
    reg.register(t0)
    reg.register(t1)

    p0.connect(t0, weight=2).connect(p1)
    t0.connect(p2).connect(t1)
    p3.connect(t1).connect(p4)

    try:
        asyncio.run(soyutnet.main(reg))
    except asyncio.exceptions.CancelledError:
        pass

    gv = reg.generate_graph()
    return gv


def feedback(N=1):
    token_ids = list(range(1, 5))
    WRITER_LABEL = 1

    async def producer(place):
        try:
            id = token_ids.pop(0)
            label = WRITER_LABEL
        except IndexError:
            id = GENERIC_ID
            label = GENERIC_LABEL
        token = (label, id)
        DEBUG(f"Produced {token}")
        return [token]

    consumed_ids = list(token_ids)

    async def consumer(place):
        nonlocal consumed_ids
        token = place.get_token(WRITER_LABEL)
        if not token:
            return
        try:
            consumed_ids.remove(token[1])
        except ValueError:
            return
        print(f"Consumed {token}")
        if not consumed_ids:
            soyutnet.terminate()

    p0 = SpecialPlace("p0", producer=producer)
    pt0 = Transition("pt0")
    p1 = Place("p1", observer_verbose=False)
    pt1 = Transition("pt1")
    p2 = Place("p2", observer_verbose=False)
    pt2 = Transition("pt2")
    p3 = SpecialPlace("p3", consumer=consumer)

    FEEDBACK_LABEL = 1
    initial_tokens = {
        GENERIC_LABEL: [GENERIC_ID] * (1 + N),
    }
    lock = Place("lock", initial_tokens=initial_tokens)

    reg = PTRegistry()
    reg.register(p0)
    reg.register(p1)
    reg.register(p2)
    reg.register(p3)
    reg.register(pt0)
    reg.register(pt1)
    reg.register(pt2)
    reg.register(lock)

    labels = [WRITER_LABEL, GENERIC_LABEL]
    (
        p0.connect(pt0, labels=labels)
        .connect(p1, labels=labels)
        .connect(pt1, labels=labels)
        .connect(p2, weight=2, labels=labels)
        .connect(pt2, weight=N + 1, labels=labels)
        .connect(p3, labels=labels[:1])
    )

    pt2.connect(lock, weight=N).connect(pt1)

    try:
        asyncio.run(soyutnet.main(reg))
    except asyncio.exceptions.CancelledError:
        pass

    print(pt0.ident(), pt0.get_no_of_times_enabled())
    print(pt1.ident(), pt1.get_no_of_times_enabled())
    print(pt2.ident(), pt2.get_no_of_times_enabled())



import asyncio
from typing import Any

from .constants import *


class Token(BaseObject):
    def __init__(
        self, label: label_t = GENERIC_LABEL, binding: Any = None, **kwargs: Any
    ) -> None:
        """
        A class that represents a labeled PTNet token which can be binded to any object.

        :param label: Label.
        :param binding: Binded object.
        """
        super().__init__(**kwargs)
        self._label: label_t = label
        """Token label"""
        self._id: id_t = GENERIC_ID
        """Unique token ID"""
        self._binding: Any = binding
        """Binded object"""

    def get_label(self) -> label_t:
        """
        Returns the label of token

        :return: Label.
        """
        return self._label

    def get_binding(self) -> Any:
        """
        Returns the object binded to the token.

        :return: Binded object.
        """
        return self._binding

    def get_id(self) -> id_t:
        """
        Returns token's unique ID.

        :return: ID.
        """
        return self._id



import sys
import asyncio

import soyutnet
from soyutnet import SoyutNet
from soyutnet.constants import GENERIC_ID, GENERIC_LABEL


def main():
    async def scheduled():
        await asyncio.sleep(1)
        soyutnet.terminate()

    net = SoyutNet()
    net.DEBUG_ENABLED = True

    LABEL = 1
    initial_tokens = {
        GENERIC_LABEL: [GENERIC_ID],
        LABEL: [1000, 990],
    }
    reg = net.PTRegistry()
    o1 = net.Observer(verbose=True)
    p1 = net.Place("p1", initial_tokens=initial_tokens, observer=o1)
    o2 = net.Observer(verbose=True)
    p2 = net.Place("p2", observer=o2)
    t1 = net.Transition("t1")
    """Define places and transitions (PTs)"""

    p1.connect(t1, labels=[GENERIC_LABEL, LABEL]).connect(
        p2, labels=[GENERIC_LABEL, LABEL]
    )
    """Connect PTs"""

    reg.register(p1)
    reg.register(p2)
    reg.register(t1)
    """Save to a list of PTs"""

    soyutnet.run(reg, extra_routines=[scheduled()])
    print("Simulation is terminated.")

    records = reg.get_merged_records()
    graph = reg.generate_graph(
        indent="  ", label_names={LABEL: "🤔", GENERIC_LABEL: "🤌"}
    )

    print("\nRecorded events:")
    {None: net.print(rec) for rec in records}
    print("\nNet graph:")
    print(graph, file=sys.stderr)

    return records, graph


if __name__ == "__main__":
    main()



import sys
import asyncio

import soyutnet
from soyutnet import SoyutNet
from soyutnet.constants import GENERIC_ID, GENERIC_LABEL, TokenType


def n_tester(n=2):
    net = SoyutNet()

    consumed_count = n + 1

    async def consumer(place):
        nonlocal consumed_count
        token: TokenType = place.get_token(GENERIC_LABEL)
        if token:
            net.DEBUG("Consumed:", token)
            consumed_count -= 1
            if consumed_count == 0:
                soyutnet.terminate()

    o1 = net.ComparativeObserver(
        expected={1: [((GENERIC_LABEL, i),) for i in range(2 * n, n - 1, -1)]},
        verbose=True,
    )
    p1 = net.Place(
        "p1", initial_tokens={GENERIC_LABEL: [GENERIC_ID] * (2 * n)}, observer=o1
    )
    t1 = net.Transition()
    p2 = net.SpecialPlace("p2", consumer=consumer)

    reg = net.PTRegistry()
    reg.register(p1)
    reg.register(t1)
    reg.register(p2)

    p1.connect(t1, weight=n).connect(p2, weight=1)
    t1.connect(p1, weight=n - 1)

    try:
        asyncio.run(soyutnet.main(reg))
    except asyncio.exceptions.CancelledError:
        print("Simulation is terminated.")
        pass

    gv = reg.generate_graph()
    return gv


if __name__ == "__main__":
    gv = n_tester(int(sys.argv[1]) + 1)
    with open("test.gv", "w") as fh:
        fh.write(gv)



import pytest
import asyncio
import random

import soyutnet
from soyutnet import SoyutNet
from soyutnet.constants import (
    GENERIC_LABEL,
    GENERIC_ID,
    INVALID_ID,
    INITIAL_ID,
    INVALID_LABEL,
)


@pytest.mark.asyncio
async def test_01():
    net = SoyutNet()
    registry = net.TokenRegistry()
    token1 = net.Token()
    token2 = net.Token()

    assert (INVALID_ID, None) == registry.get_first_entry(token1.get_label())
    registry.register(token1)
    registry.register(token2)
    assert token1.get_id() == INITIAL_ID + 1
    assert token1.get_label() == GENERIC_LABEL
    assert (token1.get_id(), token1) == registry.get_first_entry(token1.get_label())

    assert registry.get_entry_count() == 1
    assert registry.get_entry_count(INVALID_LABEL) == 0
    assert registry.pop_entry(GENERIC_LABEL, token1._id) == token1
    assert registry.pop_entry(INVALID_LABEL, random.randint(10, 1000)) == None
    assert registry.pop_entry(GENERIC_LABEL) == token2
    assert registry.pop_entry(INVALID_LABEL) == None


@pytest.mark.asyncio
async def test_02():
    net = SoyutNet()
    place = net.Place()

    assert place.get_id() == GENERIC_ID
    assert place.get_binding() is None


@pytest.mark.asyncio
async def test_03():
    net = SoyutNet()
    transition = net.Transition()

    assert transition.get_id() == GENERIC_ID
    assert transition.get_binding() is None


def test_04():
    import simple_example as e

    e.main()


def test_05():
    import simple_example_different_weight as e

    e.main_01(3, 2)


def test_06():
    import simple_example_two_input_places as e

    for i in range(100, 10000, 1000):
        e.main(i)


def test_07():
    import simple_example_two_input_places_but_different_weights as e

    e.main(w1=3, w2=2)


def test_08():
    from basic_models import co_begin

    for i in range(2, 100):
        co_begin(i)


def test_09():
    from basic_models import co_end

    for i in range(2, 100):
        co_end(i)


def test_10():
    from basic_models import sync_by_signal

    sync_by_signal()


def test_11():
    from n_tester import n_tester

    n_tester()


def test_12():
    import hashlib
    from readme_example import main

    records, graph = main()
    assert (
        hashlib.sha256(graph.encode("utf-8")).hexdigest()
        == "f66957886ed63ee512bbcfad0666501313fcd0e09f6b6a632b4def2b6675037e"
    )
    assert len(records) == 3
    assert records[0][0] == "p1"
    assert records[0][1][1:] == (((0, 1), (1, 2)), "t1")
    assert records[1][0] == "p1"
    assert records[1][1][1:] == (((0, 0), (1, 2)), "t1")
    assert records[2][0] == "p1"
    assert records[2][1][1:] == (((0, 0), (1, 1)), "t1")


def test_13():
    import hashlib
    from periodic_example import main

    for i in range(1, 100, 4):
        records = main(i)
        j = 0
        for rec in records:
            n = rec[0]
            assert (j % (i + 1) == 0 and n == "p1") or (j % (i + 1) != 0 and n == "p2")
            j += 1
